# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for Spock"""

import ast
from enum import EnumMeta
import os
import subprocess
import sys
from time import localtime
from time import strftime
from warnings import warn
import git
minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias


def convert_save_dict(clean_inner_dict, inner_val, inner_key):
    """Convert tuples in save dictionary

    *Args*:

        clean_inner_dict: inner dictionary that is clean
        inner_val: inner value
        inner_key:  inner key

    *Returns*:

        clean_inner_dict: updated with cleaned values

    """
    # Convert tuples to lists so they get written correctly
    if isinstance(inner_val, tuple):
        clean_inner_dict.update({inner_key: list(inner_val)})
    elif inner_val is not None:
        clean_inner_dict.update({inner_key: inner_val})
    return clean_inner_dict


def make_argument(arg_name, arg_type, parser):
    """Make argparser argument based on type

    Based on the type passed in handle the creation of the argparser argument so that overrides will have the correct
    behavior when set

    *Args*:

        arg_name: name for the argument
        arg_type: type of the argument
        parser: current parser

    Returns:

        parser: updated argparser

    """
    # For generic alias we take the input string and use a custom type callable to convert
    if isinstance(arg_type, _GenericAlias):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For choice enums we need to check a few things first
    elif isinstance(arg_type, EnumMeta):
        type_set = list({type(val.value) for val in arg_type})[0]
        parser.add_argument(arg_name, required=False, type=type_set)
    # For booleans we map to store true
    elif arg_type == bool:
        parser.add_argument(arg_name, required=False, action='store_true')
    # Else we are a simple base type which we can cast to
    else:
        parser.add_argument(arg_name, required=False, type=arg_type)
    return parser


def _handle_generic_type_args(val):
    return ast.literal_eval(val)


def add_info(out_dict):
    """Adds extra information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict = add_date_info(out_dict)
    out_dict = add_repo_info(out_dict)
    return out_dict


def make_blank_git(out_dict):
    """Adds blank git info

    *Args*:

        out_dict: current output dictionary

    Returns:

        out_dict: output dictionary with added git info

    """
    for key in ('BRANCH', 'COMMIT SHA', 'STATUS', 'ORIGIN'):
        out_dict.update({f'# Git {key}': 'UNKNOWN'})
    return out_dict


def add_repo_info(out_dict):
    """Adds GIT information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    try:
        # Assume we are working out of a repo
        repo = git.Repo(os.getcwd())
        # Check if we are really in a detached head state as this will fail
        if minor < 7:
            head_result = subprocess.run('git rev-parse --abbrev-ref --symbolic-full-name HEAD', stdout=subprocess.PIPE,
                                         shell=True, check=False)
        else:
            head_result = subprocess.run('git rev-parse --abbrev-ref --symbolic-full-name HEAD', capture_output=True,
                                         shell=True, check=False)
        if head_result.stdout.decode().rstrip('\n') == 'HEAD':
            out_dict = make_blank_git(out_dict)
        else:
            out_dict.update({'# Git BRANCH': repo.active_branch.name})
            out_dict.update({'# Git COMMIT SHA': repo.head.object.hexsha})
            if len(repo.untracked_files) > 0 or len(repo.head.commit.diff(None)) > 0:
                git_status = 'DIRTY'
            else:
                git_status = 'CLEAN'
            out_dict.update({'# Git STATUS': git_status})
            out_dict.update({'# Git ORIGIN': repo.remotes.origin.url})
    except git.InvalidGitRepositoryError:
        # But it's okay if we are not
        out_dict = make_blank_git(out_dict)

    return out_dict


def add_date_info(out_dict):
    """Adds date information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict.update({'# Run Date': strftime('%Y_%m_%d_%H_%M_%S', localtime())})
    return out_dict


def deep_payload_update(source, updates):
    """Deeply updates a dictionary

    Iterates through a dictionary recursively to update individual values within a possibly nested dictionary
    of dictionaries -- creates a dictionary if empty and trying to recurse

    *Args*:

        source: source dictionary
        updates: updates to the dictionary

    *Returns*:

        source: updated version of the source dictionary

    """

    for k, v in updates.items():
        if isinstance(v, dict) and v:
            source_dict = {} if source.get(k) is None else source.get(k)
            updated_dict = deep_payload_update(source_dict, v)
            if updated_dict:
                source[k] = updated_dict
        else:
            source[k] = v
    return source


def check_payload_overwrite(payload, updates, configs, overwrite=''):
    """Warns when parameters are overwritten across payloads as order will matter

    *Args*:

        payload: current payload
        payload_update: update to add to payload
        configs: config path
        overwrite: name of parent

    *Returns*:

    """
    for k, v in updates.items():
        if isinstance(v, dict) and v:
            overwrite += (k + ":")
            current_payload = {} if payload.get(k) is None else payload.get(k)
            check_payload_overwrite(current_payload, v, configs, overwrite=overwrite)
        else:
            if k in payload.keys():
                warn(f'Overriding an already set parameter {overwrite + k} from {configs}\n'
                     f'Be aware that value precedence is set by the order of the config files (last to load)...',
                     SyntaxWarning)



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import sys
import attr
from spock.backend.base import BaseBuilder


class AttrBuilder(BaseBuilder):
    """Attr specific builder

    Class that handles building for the attr backend

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """
    def __init__(self, *args, configs=None, create_save_path=False, desc='', no_cmd_line=False, **kwargs):
        super().__init__(*args, configs=configs, create_save_path=create_save_path, desc=desc,
                         no_cmd_line=no_cmd_line, **kwargs)
        for arg in self.input_classes:
            if not attr.has(arg):
                raise TypeError('*arg inputs to ConfigArgBuilder must all be class instances with attrs attributes')

    def print_usage_and_exit(self, msg=None, sys_exit=True):
        print('USAGE:')
        print(f'  {sys.argv[0]} -c [--config] config1 [config2, config3, ...]')
        print('CONFIG:')
        for attrs_class in self.input_classes:
            print('  ' + attrs_class.__name__ + ':')
            for val in attrs_class.__attrs_attrs__:
                type_string = val.metadata['base']
                # Construct the type with the metadata
                if 'optional' in val.metadata:
                    type_string = "Optional[{0}]".format(type_string)
                print(f'    {val.name}: {type_string}')
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(1)

    def _handle_arguments(self, args, class_obj):
        attr_name = class_obj.__name__
        class_names = [val.__name__ for val in self.input_classes]
        # Handle repeated classes
        if attr_name in class_names and attr_name in args and isinstance(args[attr_name], list):
            fields = self._handle_repeated(args[attr_name], attr_name, class_names)
        # Handle non-repeated classes
        else:
            fields = {}
            for val in class_obj.__attrs_attrs__:
                # Check if namespace is named and then check for key -- checking for local class def
                if attr_name in args and val.name in args[attr_name]:
                    fields[val.name] = self._handle_nested_class(args, args[attr_name][val.name], class_names)
                # If not named then just check for keys -- checking for global def
                elif val.name in args:
                    fields[val.name] = self._handle_nested_class(args, args[val.name], class_names)
                # Check for special keys to set
                if 'special_key' in val.metadata and val.metadata['special_key'] is not None:
                    if val.name in args:
                        self.save_path = args[val.name]
                    elif val.default is not None:
                        self.save_path = val.default
        return fields

    def _handle_repeated(self, args, check_value, class_names):
        """Handles repeated classes as lists

        *Args*:

            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            list of input_class[match)idx[0]] types filled with repeated values

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        return [self.input_classes[match_idx[0]](**val) for val in args]

    def _handle_nested_class(self, args, check_value, class_names):
        """Handles passing another class to the field dictionary

        *Args*:
            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            either the check_value or the necessary class

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        # If so then create the needed class object by unrolling the args to **kwargs and return it
        if len(match_idx) > 0:
            if len(match_idx) > 1:
                raise ValueError('Match error -- multiple classes with the same name definition')
            else:
                if args.get(self.input_classes[match_idx[0]].__name__) is None:
                    raise ValueError(f'Missing config file definition for the referenced class '
                                     f'{self.input_classes[match_idx[0]].__name__}')
                current_arg = args.get(self.input_classes[match_idx[0]].__name__)
                if isinstance(current_arg, list):
                    class_value = [self.input_classes[match_idx[0]](**val) for val in current_arg]
                else:
                    class_value = self.input_classes[match_idx[0]](**current_arg)
            return_value = class_value
        # else return the expected value
        else:
            return_value = check_value
        return return_value



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles payloads from markup files"""

from itertools import chain
from spock.backend.attr.utils import convert_to_tuples
from spock.backend.attr.utils import get_type_fields
from spock.backend.attr.utils import deep_update
from spock.backend.base import BasePayload


class AttrPayload(BasePayload):
    """Handles building the payload for attrs backend

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class

    """
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        """Call to allow self chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            Payload: instance of self

        """
        return AttrPayload()

    @staticmethod
    def _update_payload(base_payload, input_classes, payload):
        # Get basic args
        attr_fields = {attr.__name__: [val.name for val in attr.__attrs_attrs__] for attr in input_classes}
        # Class names
        class_names = [val.__name__ for val in input_classes]
        # Parse out the types if generic
        type_fields = get_type_fields(input_classes)
        for keys, values in base_payload.items():
            # check if the keys, value pair is expected by the attr class
            if keys != 'config':
                # Dict infers that we are overriding a global setting in a specific config
                if isinstance(values, dict):
                    # we're in a namespace
                    # Check for incorrect specific override of global def
                    if keys not in attr_fields:
                        raise TypeError(f'Referring to a class space {keys} that is undefined')
                    for i_keys in values.keys():
                        if i_keys not in attr_fields[keys]:
                            raise ValueError(f'Provided an unknown argument named {keys}.{i_keys}')
                else:
                    # Check if the key is actually a reference to another class
                    if keys in class_names:
                        if isinstance(values, list):
                            # Check for incorrect specific override of global def
                            if keys not in attr_fields:
                                raise TypeError(f'Referring to a class space {keys} that is undefined')
                            # We are in a repeated class def
                            # Raise if the key set is different from the defined set (i.e. incorrect arguments)
                            key_set = set(list(chain(*[list(val.keys()) for val in values])))
                            for i_keys in key_set:
                                if i_keys not in attr_fields[keys]:
                                    raise ValueError(f'Provided an unknown argument named {keys}.{i_keys}')
                        # Chain all the values from multiple spock classes into one list
                        elif keys not in list(chain(*attr_fields.values())):
                            raise ValueError(f'Provided an unknown argument named {keys}')
                    # Chain all the values from multiple spock classes into one list
                    elif keys not in list(chain(*attr_fields.values())):
                        raise ValueError(f'Provided an unknown argument named {keys}')
            if keys in payload and isinstance(values, dict):
                payload[keys].update(values)
            else:
                payload[keys] = values
        tuple_payload = convert_to_tuples(payload, type_fields, class_names)
        payload = deep_update(payload, tuple_payload)
        return payload
